/*
 *    WMAcpiLoad - A dockapp to monitor ACPI status
 *    Copyright (C) 2002  Thomas Nemeth <tnemeth@free.fr>
 *
 *    Patch by Alan Carriou <cariou_alan@yahoo.fr>
 *
 *    Based on work by Seiichi SATO <ssato@sh.rim.or.jp>
 *    Copyright (C) 2001,2002  Seiichi SATO <ssato@sh.rim.or.jp>
 *    and on work by Mark Staggs <me@markstaggs.net>
 *    Copyright (C) 2002  Mark Staggs <me@markstaggs.net>

 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.

 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.

 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <errno.h>                              /* needed by errno */
#include <signal.h>
#include "dockapp.h"
#include "backlight_on.xpm"
#include "backlight_off.xpm"
#include "parts.xpm"

#ifdef linux
# include <sys/stat.h>
# include <dirent.h>                                /* needed by scandir() */
# define ACPIDEV "/proc/acpi/info"
# ifndef ACPI_32_BIT_SUPPORT
#  define ACPI_32_BIT_SUPPORT      0x0002
# endif
#endif

#define FREE(data) {if (data) free (data); data = NULL;}

#define SIZE        58
#define MAXSTRLEN   512
#define WINDOWED_BG ". c #AEAAAE"
#define MAX_HISTORY 16
#define CPUNUM_NONE -1
/* default number of seconds between updates */
#define DEFAULT_UPDATE_TIME      1
/* default battery status under which alarm mode is set */
#define DEFAULT_ALARM_BAT_LEVEL 20

typedef enum { LIGHTOFF, LIGHTON } light;
typedef enum { On_line, Off_line } ac_power;
typedef enum { Charging, Discharging, Unknown } bat_status;

/* char buffer to contain file name, such as "BAT1", "TZ1"...
 *
 * The "ACPI Specification" (version 2.0c) tells that the naming conventions
 * (a.k.a. "ACPI Namespace") imposes 32-bits-long names. So a char[4] is,
 * theoretically, enough. But many bioses (acutally, DSDTs) are not 100% spec-
 * compliant, so i'd rather keep some margin.
 *
 * Moreover, the names we use here are not taken directly from the machine's
 * acpi BIOS, but through the OS interface, which can be buggy, or follow a
 * different naming convention (e.g. a manually fixed DSDT).
 *
 * So i guess it is better (I mean, more likely to work) to follow the
 * hypothesis that the names do not follow any restrictive spec, and to use
 * whatever names we get. So i've chosen the maximum file name length. (char*
 * caused some trouble within parse_dir)
 */
typedef char device_name[NAME_MAX + 1];

/* structure to hold information about ONE ac adapter
 *
 * Managing several acad's can be a real need. The "ACPI Specification 2.0c"
 * (section 11.4) gives the example of a laptop plugged in a docking station,
 * each one having its own ACad, which means that 2 AC adapters are listed in
 * the global acpi namespace.
 *
 * I do not know yet how it managed by any OS implementation ; neither do I
 * know about the possibility of an ACad to de declared as "on-line" when
 * another is "off-line". I haven't read the whole 518 pages of the acpi spec,
 * and i do not think i'll have the time for this, so I would apreciate any
 * help !
 *
 * Meanwhile, each AC_adapter holds a 'status' data, and we do not check
 * whether all these 'status' are consistent. Moreover, we monitor only the
 * first acad (no next_acad() function).
 *
 * As far as the GUI is concerned, the icon "ac power on-line" is displayed
 * when the status of the monitored ac adapter (see the strucutre ACPI_Infos)
 * is set to "On_line".
 */
typedef struct {
    device_name name;        /* device name given by OS acpi interface */
    ac_power    status;
/*  char        *alias;*/    /* human understandable alias */
} AC_adapter;

/* structure to hold information about ONE battery
 *
 * "percentage" is stored into de structure, so that we only have to calculate
 * it once when the remaining capacity is stilll (means battery plugged and at
 * full load).
 *
 * The values of currcap, remain and percentage when the battery is unplugged
 * should be 1, -1, -1 (until the code is adapted to manage this situation)
 */
typedef struct {
    device_name  name;
    int          plugged;    /* 1 when a battery is plugged in the slot, else 0 */
    float        currcap;    /* "last full capacity" */
    float        remain;     /* remaining capacity */
    int          percentage; /* load percentage = remain/currcap * 100 */
    bat_status   status;     /* charging / discharging / unknown */
} Battery;

/* structure to hold information about ONE thermal sensor ( = "Thermal Zone"
 * in ACPI naming standard).
 *
 * temp is set to 0 when unconsitent date is read.
 */
typedef struct {
    device_name  name;
    unsigned int temp;        /* the temperature read by the sensor */
} Therm_Zone;

/* structure to hold every acpi information */
typedef struct {
  /*const char driver_version[10];*/ /* unused */
    AC_adapter *ac;           /* list of acad's */
    AC_adapter *ac_active;    /* monitored acad */
    int        ac_num;        /* number of ac adapters */

    Battery    *bat;          /* list of battery slots */
    Battery    *bat_active;   /* monitored battery */
    int        bat_num;       /* number of battery slots */

    Therm_Zone *thermal;      /* list of thermal zones */
    Therm_Zone *tz_active;    /* monitored thermal zone */
    int        tz_num;        /* number of thermal zones */
} AcpiInfos;


Pixmap pixmap;
Pixmap backdrop_on;
Pixmap backdrop_off;
Pixmap parts;
Pixmap mask;
static char     *display_name     = "";   /* which X Window display to use */
static char     *light_color      = NULL; /* back-light color */
static unsigned update_interval   = DEFAULT_UPDATE_TIME;
static light    backlight         = LIGHTOFF;
static unsigned switch_authorized = True;
static unsigned alarm_level       = DEFAULT_ALARM_BAT_LEVEL;
/* default temperature above which alarm mode is set */
static unsigned alarm_level_temp  = 70;

static char     *notif_cmd        = NULL;/* executed when entering alarm mode*/
static char     *suspend_cmd      = NULL; /* unused */
static char     *standby_cmd      = NULL; /* unused */

/* non-null value causes debug messages on standard output */
static int      _verbose          = 0;

/* static variable to hold every useful acpi information about the machine */
static AcpiInfos cur_acpi_infos;

/* prototypes */
static void update(void);
static void switch_light(void);
static void draw_timedigit(AcpiInfos infos);
static void draw_pcdigit(AcpiInfos infos);
static void draw_statusdigit(AcpiInfos infos);
static void draw_pcgraph(AcpiInfos infos);
static void parse_arguments(int argc, char **argv);
static void print_help(char *prog);
static void acpi_getinfos(AcpiInfos *infos);
static int  acpi_exists(void);
static int  my_system (char *cmd);
static void next_thermal(AcpiInfos *infos);
static void next_battery(AcpiInfos *infos);
#ifdef linux
int acpi_read(AcpiInfos *i);
void strncat3(char **dest, int size, char *dir, char *name, char *file);
void init_stats(AcpiInfos *k);
void check_battery_plugged (Battery *bat);
void get_battery_currcap (Battery *bat);
void init_stats_bat(AcpiInfos *k);
void init_stats_thermal(AcpiInfos *k);
void init_stats_ac(AcpiInfos *k);
int parse_dir (char *dir, device_name **list);
#endif


int count;         /* what does it count ? */


int main(int argc, char **argv) {
    XEvent   event;
    XpmColorSymbol colors[2] = { {"Back0", NULL, 0}, {"Back1", NULL, 0} };
    int      ncolor = 0;
    struct   sigaction sa;

    sa.sa_handler = SIG_IGN;
#ifdef SA_NOCLDWAIT
    sa.sa_flags = SA_NOCLDWAIT;
#else
    sa.sa_flags = 0;
#endif
    sigemptyset(&sa.sa_mask);
    sigaction(SIGCHLD, &sa, NULL);

    /* Parse CommandLine */
    parse_arguments(argc, argv);

    /* Check for ACPI support */
    if (!acpi_exists()) {
#ifdef linux
        fprintf(stderr, "No ACPI support in kernel\n");
#else
        fprintf(stderr, "Unable to access ACPI info\n");
#endif
        exit(1);
    }

    /* Initialize ACPI information (mainly names and number of available
     * devices).
     */
    init_stats(&cur_acpi_infos);

    dockapp_open_window(display_name, PACKAGE, SIZE, SIZE, argc, argv);
    dockapp_set_eventmask(ButtonPressMask);

    if (light_color) {
        colors[0].pixel = dockapp_getcolor(light_color);
        colors[1].pixel = dockapp_blendedcolor(light_color, -24, -24, -24, 1.0);
        ncolor = 2;
    }

    /* change raw xpm data to pixmap */
    if (dockapp_iswindowed)
        backlight_on_xpm[1] = backlight_off_xpm[1] = WINDOWED_BG;

    if (!dockapp_xpm2pixmap(backlight_on_xpm, &backdrop_on, &mask, colors, ncolor)) {
        fprintf(stderr, "Error initializing backlit background image.\n");
        exit(1);
    }
    if (!dockapp_xpm2pixmap(backlight_off_xpm, &backdrop_off, NULL, NULL, 0)) {
        fprintf(stderr, "Error initializing background image.\n");
        exit(1);
    }
    if (!dockapp_xpm2pixmap(parts_xpm, &parts, NULL, colors, ncolor)) {
        fprintf(stderr, "Error initializing parts image.\n");
        exit(1);
    }

    /* shape window */
    if (!dockapp_iswindowed) dockapp_setshape(mask, 0, 0);
    if (mask) XFreePixmap(display, mask);

    /* pixmap : draw area */
    pixmap = dockapp_XCreatePixmap(SIZE, SIZE);

    /* Initialize pixmap */
    if (backlight == LIGHTON)
        dockapp_copyarea(backdrop_on, pixmap, 0, 0, SIZE, SIZE, 0, 0);
    else
        dockapp_copyarea(backdrop_off, pixmap, 0, 0, SIZE, SIZE, 0, 0);

    dockapp_set_background(pixmap);
    dockapp_show();

    /* get acpi information (temperature, battery status...) and displays them
     * for the first time.
     */
    update();

    /* Main loop */
    while (1) {
        if (dockapp_nextevent_or_timeout(&event, update_interval * 1000)) {
            /* Next Event */
            switch (event.type) {
            case ButtonPress:
                switch (event.xbutton.button) {
                case 1: switch_light(); break;
                case 2: next_thermal(&cur_acpi_infos); update(); break;
                case 3: next_battery(&cur_acpi_infos); update(); break;
                default: break;
                }
                    break;
            default: break;
            }
        } else {
            /* Time Out */
            update();
        }
    }

    return 0;
}

void init_stats(AcpiInfos *k) {
    init_stats_thermal(k);
    init_stats_ac(k);
    init_stats_bat(k);

    /*
     * This check is done only once, so that the user can know exactly which
     * devices are detected by the program, and which are not (using verbose
     * mode). If it was done in each of the functions above, it could exit
     * from initstats_thermal, and the user wouldn't know if the problem is
     * only about thermal zones, or about something else.
     *
     * We assume that error messages were given by the respective function, so
     * we here only have to exit.
     */
    if (k->ac_num == 0 || k->bat_num == 0 || k->tz_num == 0)
        exit(1);
}

#ifdef linux
/*
 * Parse /proc/acpi/thermal_zone, to get the number and names of thermal
 * sensors.
 */
void init_stats_thermal(AcpiInfos *k) {
    int         i;
    device_name *list;

    /* We first have to get the number and names of thermal zones */
    k->tz_num = parse_dir("/proc/acpi/thermal_zone", &list);

    if (k->tz_num <= 0) {
        fprintf(stderr, "No thermal sensor found\n");
        k->tz_num = 0;
        k->thermal = NULL;
    } else {
        if (_verbose) {
            printf ("I have found %s thermal sensor%s:",
                    (k->tz_num == 1) ? "this" : "those",
                    (k->tz_num == 1) ? "" : "s");
            for (i = 0; i < k->tz_num ; i++)
                printf(" %s", list[i]);
            printf("\n");
        }

        /* Then we can create the Therm_Zone structures */
        k->thermal = (Therm_Zone *) malloc(sizeof(Therm_Zone)*k->tz_num);
        if (!k->thermal) {
            fprintf(stderr, PACKAGE ": cannot allocate memory "
                            "for thermal zones information\n");
            perror("");
            exit(1);
        }
        for (i = 0; i < k->tz_num; i++) {
            strcpy (k->thermal[i].name, list[i]);
            k->thermal[i].temp = 0;         /* default value */
        }
    }

    k->tz_active = k->thermal;
}

/*
 * Parse /proc/acpi/ac_adapter, to get the number and names of the AC adapters.
 */
void init_stats_ac (AcpiInfos *k)
{
    int         i;
    device_name *list;

    /* We first have to get the number and names of acads */
    k->ac_num = parse_dir("/proc/acpi/ac_adapter", &list);
    if (k->ac_num <= 0) {
        fprintf(stderr, "No AC adapter found\n");
        k->ac_num = 0;
        k->ac = NULL;
    } else {
        if (_verbose) {
            printf ("I have found %s AC adapter%s:",
                (k->ac_num == 1) ? "this" : "those",
                (k->ac_num == 1) ? "" : "s");
            for (i = 0; i < k->ac_num ; i++)
                printf(" %s", list[i]);
            printf("\n");
        }

        /* Then we can create the AC_adapter structures */
        k->ac = (AC_adapter *) malloc(sizeof(AC_adapter)*k->ac_num);
        if (!k->ac) {
            fprintf(stderr, PACKAGE ": cannot allocate memory "
                            "for AC adapters information\n");
            perror("");
            exit(1);
        }
        for (i = 0; i < k->ac_num; i++) {
            strcpy(k->ac[i].name, list[i]);
            k->ac[i].status = Off_line;           /* default value */
        }
    }
    k->ac_active = k->ac;
}

/*
 * Used to get some file names as "/proc/acpi/battery/"+"C11C"+"/state"
 *
 * (*dest) should be free'd before calling this function.
 */
void strncat3(char **dest, int size, char *dir, char *name, char *file)
{
    *dest = (char *)malloc(sizeof(char)*size);

    if (*dest == NULL) {
        fprintf(stderr, PACKAGE ": cannot allocate memory for file name\n");
        perror("");
        exit(1);
    }

    strncpy(*dest, dir, size - 1);
    (*dest)[size - 1] = '\0';

    strncat(*dest, name, size - strlen(*dest) - 1);
    (*dest)[size - 1] = '\0';

    strncat(*dest, file, size - strlen(*dest) - 1);
    (*dest)[size - 1] = '\0';
}

void check_battery_plugged (Battery *bat)
{
    int plugged = 0;
    FILE *fd;
    char buf[512];
    char *ptr;
    char *filename;

    strncat3 (&filename, 256, "/proc/acpi/battery/", bat->name, "/info");

    if ((fd = fopen(filename, "r"))) {
        fread(buf, sizeof buf, 1, fd);
        fclose(fd);

        /* is the battery plugged in ? */
        if ((ptr = strstr(buf, "present:"))) {
            ptr += 25;
            if (*ptr == 'y') {
                plugged = 1;
            } else if (*ptr != 'n') {
                /* "present" is neither "yes" nor "no" */
                fprintf(stderr, "Unconsistent data about battery %s\n",
                                bat->name);
            }
        } else {       /* the info file does not contain the line "present" */
            fprintf(stderr, "Unconsistent data about battery %s\n",
                            bat->name);
        }
    } else {
        /* the battery "bat" has been found (by scandir), but does not have any
         * "info" file */
        fprintf(stderr, "Unconsistent data about battery %s\n", bat->name);
    }
    free(filename);

    bat->plugged = plugged;
}

void get_battery_currcap (Battery *bat)
{
    float currcap = 1;  /* default value */
    FILE  *fd;
    char  *filename;
    char  buf[512];
    char  *ptr;

    strncat3(&filename, 512, "/proc/acpi/battery/", bat->name, "/info");

    if ((fd = fopen(filename, "r"))) {
        fread(buf, sizeof buf, 1, fd);
        fclose(fd);

        if ((ptr = strstr(buf, "last full capacity:"))) {
            ptr += 25;
            if (sscanf(ptr, "%d", &currcap) == 0) {
                /* sscanf failed */
                fprintf(stderr, "Unconsistent data about battery %s\n",
                                bat->name);
                currcap = 1;
            }
        } else {
            /* the info file does not contain any "last full capacity" */
            fprintf(stderr, "Battery %s unplugged or unconsistent data\n",
                            bat->name);
        }
    } else {
        /* The battery does not have any info file, although having been detec-
         * ted by scandir.
         */
        fprintf(stderr, "Unconsistent data about battery %s\n", bat->name);
    }
    free(filename);

    bat->currcap = currcap;
}

/*
 * Parse /proc/acpi/battery, to get information about battery slots (number and
 * names).
 */
void init_stats_bat (AcpiInfos *k)
{
    int         i;
    device_name *list;

    /* We first have to get the number and names of battery slots */
    k->bat_num = parse_dir("/proc/acpi/battery", &list);
    if (k->bat_num <= 0) {
        fprintf(stderr, "No battery found\n");
        k->bat_num = 0;
        k->bat = NULL;
        k->bat_active = NULL;
        return;
    }

    if (_verbose) {
        printf ("I have found %s battery slot%s:",
                (k->bat_num == 1) ? "this" : "those",
                (k->bat_num == 1) ? "" : "s");
        for (i = 0; i < k->bat_num ; i++)
            printf(" %s", list[i]);
        printf("\n");
    }

    k->bat = (Battery *) malloc(sizeof(Battery)*k->bat_num);
    if (!k->bat) {
        fprintf(stderr, PACKAGE ": cannot allocate memory "
                        "for battery information\n");
        perror("");
        exit(1);
    }

    /* Checks if the batteries are plugged in, and get info about full
     * battery charge (when possible), one battery after the other
     */
    for (i = 0; i < k->bat_num; i++) {
        strcpy(k->bat[i].name, list[i]);

        check_battery_plugged(&k->bat[i]);
        if (_verbose) {
            if (k->bat[i].plugged) {
                printf("Battery %s is plugged in\n", k->bat[i].name);
            } else {
                printf("Battery %s is not plugged in\n", k->bat[i].name);
            }
        }

        if (k->bat[i].plugged) {
            get_battery_currcap(&k->bat[i]);
        } else {
            k->bat[i].currcap = 1;
        }
        k->bat[i].remain = -1;
        k->bat[i].percentage = -1;
    }
    k->bat_active = k->bat;
}
#endif   /* linux */

/* called by timer */
static void update(void)
{
    static light pre_backlight;
    static Bool in_alarm_mode = False;

    /* get current system state */
    acpi_getinfos(&cur_acpi_infos);

    /* alarm mode */
    if ((cur_acpi_infos.bat_active->percentage < alarm_level)
            || (cur_acpi_infos.tz_active->temp > alarm_level_temp)) {
        if (!in_alarm_mode) {
            in_alarm_mode = True;
            pre_backlight = backlight;
            my_system(notif_cmd);
        }
        if ( (switch_authorized) ||
                ( (! switch_authorized) && (backlight != pre_backlight) ) ) {
            switch_light();
            return;
        }
    } else {
        if (in_alarm_mode) {
            in_alarm_mode = False;
            if (backlight != pre_backlight) {
                switch_light();
                return;
            }
        }
    }

    /* all clear */
    if (backlight == LIGHTON)
        dockapp_copyarea(backdrop_on, pixmap, 0, 0, 58, 58, 0, 0);
    else
        dockapp_copyarea(backdrop_off, pixmap, 0, 0, 58, 58, 0, 0);

    /* draw digit */
    draw_timedigit(cur_acpi_infos);
    draw_pcdigit(cur_acpi_infos);
    draw_statusdigit(cur_acpi_infos);
    draw_pcgraph(cur_acpi_infos);

    /* show */
    dockapp_copy2window(pixmap);
}


/* called when left mouse button pressed */
static void switch_light(void) {
    switch (backlight) {
        case LIGHTOFF:
            backlight = LIGHTON;
            dockapp_copyarea(backdrop_on, pixmap, 0, 0, 58, 58, 0, 0);
            break;
        case LIGHTON:
            backlight = LIGHTOFF;
            dockapp_copyarea(backdrop_off, pixmap, 0, 0, 58, 58, 0, 0);
            break;
    }

    /* redraw digit */
    acpi_getinfos(&cur_acpi_infos);
    draw_timedigit(cur_acpi_infos);
    draw_pcdigit(cur_acpi_infos);
    draw_statusdigit(cur_acpi_infos);
    draw_pcgraph(cur_acpi_infos);

    /* show */
    dockapp_copy2window(pixmap);
}

/* Displays the temperature of the active thermal sensor, as a two-digit
 * number.
 *
 * FIX: shouldn't we change the function name ?
 */
static void draw_timedigit(AcpiInfos infos) {
    int y = 0;
/*  int time_left, hour_left, min_left;*/

    if (backlight == LIGHTON) y = 20;

    dockapp_copyarea(parts, pixmap,
        (infos.tz_active->temp / 10)*10, y, 10, 20, 5, 7);
    dockapp_copyarea(parts, pixmap,
        (infos.tz_active->temp % 10)*10, y, 10, 20, 17, 7);
    /*hour_left = time_left / 60;
    min_left  = time_left % 60;
    dockapp_copyarea(parts, pixmap, (hour_left / 10) * 10, y, 10, 20,  5, 7);
    dockapp_copyarea(parts, pixmap, (hour_left % 10) * 10, y, 10, 20, 17, 7);
    dockapp_copyarea(parts, pixmap, (min_left / 10)  * 10, y, 10, 20, 32, 7);
    dockapp_copyarea(parts, pixmap, (min_left % 10)  * 10, y, 10, 20, 44, 7);
    */
}

static void draw_pcdigit(AcpiInfos infos) {
    int xd = 0;
    if (backlight == LIGHTON) xd = 50;

    if (infos.bat_active->plugged) {
        int v100, v10, v1;
        int num = infos.bat_active->percentage;

        if (num < 0)  num = 0;

        v100 = num / 100;
        v10  = (num - v100 * 100) / 10;
        v1   = (num - v100 * 100 - v10 * 10);

        /* draw digit */
        dockapp_copyarea(parts, pixmap, v1 * 5 + xd, 40, 5, 9, 17, 45);
        if (v10 != 0)
            dockapp_copyarea(parts, pixmap, v10 * 5 + xd, 40, 5, 9, 11, 45);
        if (v100 == 1) {
            dockapp_copyarea(parts, pixmap, 5 + xd, 40, 5, 9, 5, 45);
            dockapp_copyarea(parts, pixmap, 0 + xd, 40, 5, 9, 11, 45);
        }
    } else {
        /* battery unplugged: display "---%" */
        dockapp_copyarea(parts, pixmap, 10 + xd, 49, 5, 9,  5, 45);
        dockapp_copyarea(parts, pixmap, 10 + xd, 49, 5, 9, 11, 45);
        dockapp_copyarea(parts, pixmap, 10 + xd, 49, 5, 9, 17, 45);
    }
}


static void draw_statusdigit(AcpiInfos infos) {
    int xd = 0;
    int y = 31;

    if (backlight == LIGHTON) {
        y = 40;
        xd = 50;
    }

    /* draw digit */
    if (infos.bat_active->plugged) {
        if (infos.bat_active->status == Charging)
            dockapp_copyarea(parts, pixmap, 100, y, 4, 9, 41, 45);
    }
    else {
        /* shows '-' as charging state */
        dockapp_copyarea(parts, pixmap, 10 + xd, 49, 5, 9, 41, 45);
    }

    if (infos.ac_active->status == On_line)
        dockapp_copyarea(parts, pixmap, 0 + xd, 49, 5, 9, 34, 45);
    else
        dockapp_copyarea(parts, pixmap, 5 + xd, 49, 5, 9, 48, 45);
}

static void draw_pcgraph(AcpiInfos infos) {
    int xd = 100;
    int nb;
    int num = infos.bat_active->percentage / 6.25 ;

    if (num < 0) num = 0;

    if (backlight == LIGHTON) xd = 102;

    /* draw digit */
    for (nb = 0 ; nb < num ; nb++)
        dockapp_copyarea(parts, pixmap, xd, 0, 2, 9, 6 + nb * 3, 33);
}


static void parse_arguments(int argc, char **argv) {
    int i;
    int integer;
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h")) {
            print_help(argv[0]), exit(0);
        } else if (!strcmp(argv[i], "--version") || !strcmp(argv[i], "-v")) {
            printf("%s version %s\n", PACKAGE, VERSION), exit(0);
        } else if (!strcmp(argv[i], "--display") || !strcmp(argv[i], "-d")) {
            display_name = argv[i + 1];
            i++;
        } else if (!strcmp(argv[i], "--backlight") || !strcmp(argv[i], "-bl")) {
            backlight = LIGHTON;
        } else if (!strcmp(argv[i], "--light-color") || !strcmp(argv[i], "-lc")) {
            light_color = argv[i + 1];
            i++;
        } else if (!strcmp(argv[i], "--interval") || !strcmp(argv[i], "-i")) {
            if (argc == i + 1)
                fprintf(stderr, "%s: error parsing argument for option %s\n",
                        argv[0], argv[i]), exit(1);
            if (sscanf(argv[i + 1], "%i", &integer) != 1)
                fprintf(stderr, "%s: error parsing argument for option %s\n",
                        argv[0], argv[i]), exit(1);
            if (integer < 1)
                fprintf(stderr, "%s: argument %s must be >=1\n",
                        argv[0], argv[i]), exit(1);
            update_interval = integer;
            i++;
        } else if (!strcmp(argv[i], "--alarm") || !strcmp(argv[i], "-a")) {
            if (argc == i + 1)
                    fprintf(stderr, "%s: error parsing argument for option %s\n",
                            argv[0], argv[i]), exit(1);
            if (sscanf(argv[i + 1], "%i", &integer) != 1)
                fprintf(stderr, "%s: error parsing argument for option %s\n",
                        argv[0], argv[i]), exit(1);
            if ( (integer < 0) || (integer > 100) )
                fprintf(stderr, "%s: argument %s must be >=0 and <=100\n",
                        argv[0], argv[i]), exit(1);
            alarm_level = integer;
            i++;
        } else if (!strcmp(argv[i], "--windowed") || !strcmp(argv[i], "-w")) {
            dockapp_iswindowed = True;
        } else if (!strcmp(argv[i], "--broken-wm") || !strcmp(argv[i], "-bw")) {
            dockapp_isbrokenwm = True;
        } else if (!strcmp(argv[i], "--notify") || !strcmp(argv[i], "-n")) {
            notif_cmd = argv[i + 1];
            i++;
        } else if (!strcmp(argv[i], "--suspend") || !strcmp(argv[i], "-s")) {
            suspend_cmd = argv[i + 1];
            i++;
        } else if (!strcmp(argv[i], "--standby") || !strcmp(argv[i], "-S")) {
            standby_cmd = argv[i + 1];
            i++;
        } else if (!strcmp(argv[i], "--verbose") || !strcmp(argv[i], "-V")) {
            _verbose = 1;
        } else {
            fprintf(stderr, "%s: unrecognized option '%s'\n", argv[0], argv[i]);
            print_help(argv[0]), exit(1);
        }
    }
}


static void print_help(char *prog)
{
    printf("Usage : %s [OPTIONS]\n"
    "%s - Window Maker acpi monitor dockapp\n"
    "  -d,  --display <string>        display to use\n"
    "  -bl, --backlight               turn on back-light\n"
    "  -lc, --light-color <string>    back-light color(rgb:6E/C6/3B is "
                                      "default)\n"
    "  -i,  --interval <number>       number of secs between updates (%d is "
                                      "default)\n"
    "  -a,  --alarm <number>          low battery level when to raise alarm "
                                      "(%d is default)\n"
    "  -h,  --help                    show this help text and exit\n",
        prog, prog, DEFAULT_UPDATE_TIME, DEFAULT_ALARM_BAT_LEVEL);
    printf(
    "  -v,  --version                 show program version and exit\n"
    "  -V,  --verbose                 display debug messages on standard "
                                      "output\n"
    "  -w,  --windowed                run the application in windowed mode\n"
    "  -bw, --broken-wm               activate broken window manager fix\n"
    "  -n,  --notify <string>         command to launch when alarm is on\n"
    "  -s,  --suspend <string>        set command for acpi suspend\n"
    "  -S,  --standby <string>        set command for acpi standby\n");
    /* OPTIONS SUPP :
     *  ? -f, --file    : configuration file
     */

    /* 'gcc -pedantic' tells that the message is too long to be printf-ed at
     * once by ISO C89 compilers...
     */
}


static void acpi_getinfos(AcpiInfos *infos) {
    if (
#if defined(linux) || defined(solaris)
    (acpi_read(infos))
#else
# ifdef freebsd
    (acpi_read(&temp_info))
# endif
#endif
     ) {
        fprintf(stderr, "Cannot read ACPI information.\n");
    }
}


int acpi_exists(void)
{
    return (access(ACPIDEV, R_OK)) ? 0 : 1;
}

/*
 * Calls sh to execute an external command (alarm, suspend or standby).
 */
static int my_system (char *cmd) {
    int           pid;
    extern char **environ;

    if (cmd == 0) return 1;
    pid = fork ();
    if (pid == -1) return -1;
    if (pid == 0) {
        pid = fork ();
        if (pid == 0) {
            char *argv[4];
            argv[0] = "sh";
            argv[1] = "-c";
            argv[2] = cmd;
            argv[3] = 0;
            execve ("/bin/sh", argv, environ);
            exit (0);
        }
        exit (0);
    }
    return 0;
}

/*
 * Called within the main loop to ++ "tz_active".
 */
void next_thermal(AcpiInfos *infos)
{
    if (infos->tz_active == infos->thermal + infos->tz_num - 1)
        infos->tz_active = infos->thermal;
    else
        infos->tz_active++;

    if (_verbose)
        printf ("Now monitoring thermal zone %s.\n", infos->tz_active->name);
}

/*
 * Called within the main loop to ++ "bat_active".
 */
void next_battery(AcpiInfos *infos)
{
    if (infos->bat_active == infos->bat + infos->bat_num - 1)
        infos->bat_active = infos->bat;
    else
        infos->bat_active++;

    if (_verbose)
        printf ("Now monitoring battery %s.\n", infos->bat_active->name);
}

#ifdef linux
/* Reads the informations (temperature, battery status...) from /proc/acpi. It
 * is called within update().
 */
int acpi_read(AcpiInfos *i)
{
    FILE *fd;
    int  retcode = 0;
    int  tmp;
    char buf[512];
    char *ptr;
    char *file;
    char stat;

    /*
    if ((fd = fopen("/proc/acpi/thermal_zone/THRM/state", "r"))) {
        bzero(buf, 512);
        fscanf(fd, "state: %s", buf);
        fclose(fd);
        if (strstr(buf, "ok")) i->thermal_state=1;
    }*/

    /* get acpi thermal info */
    strncat3(&file, 256, "/proc/acpi/thermal_zone/", i->tz_active->name,
               "/temperature");

    if ((fd = fopen(file, "r"))) {
        if (fscanf(fd, "temperature: %d", &tmp) == 1) {
            i->tz_active->temp = tmp;
        } else {
            /* fscanf failed */
            fprintf(stderr, "Unconsistent data about thermal zone %s\n",
                            i->tz_active->name);
            i->tz_active->temp = 0;
        }
        fclose(fd);
    }
    free(file);

    /* get acpi AC info */
    strncat3(&file, 256, "/proc/acpi/ac_adapter/", i->ac_active->name,
               "/state");
    if ((fd = fopen(file, "r"))) {
        bzero(buf, sizeof buf);
        fscanf(fd, "state: %s", buf);
        fclose(fd);
        if(strstr(buf, "on-line") != NULL)  i->ac_active->status = On_line;
        if(strstr(buf, "off-line") != NULL) i->ac_active->status = Off_line;
    }
    free(file);

    if (!i->bat_active->plugged) {
        /* check whether the battery has been plugged in since last update */
        check_battery_plugged (i->bat_active);

        if (i->bat_active->plugged) {
            if (_verbose) {
                printf("Battery %s has been plugged in\n",i->bat_active->name);
            }
            get_battery_currcap (i->bat_active);
        }
    } else {
        /* check whether the battery has been plugged out since last update */
        check_battery_plugged (i->bat_active);
        if ((!i->bat_active->plugged) && (_verbose)) {
            printf("Battery %s has been plugged out\n", i->bat_active->name);
        }
    }

    /* get acpi battery info, if plugged */
    if (i->bat_active->plugged) {
        strncat3(&file, 512, "/proc/acpi/battery/", i->bat_active->name,
                "/state");

        if ((fd = fopen(file, "r"))) {
            bzero(buf, sizeof buf);
            fread(buf, sizeof buf, 1, fd);
            fclose(fd);

            if ((ptr = strstr(buf, "charging state:"))) {
                stat = *(ptr + 25);
                switch (stat) {
                case 'd': i->bat_active->status = Discharging; break;
                case 'c': i->bat_active->status = Charging;    break;
                case 'u': i->bat_active->status = Unknown;     break;
                }
            } else {
                i->bat_active->status = Unknown;
            }

            if ((ptr = strstr(buf, "remaining capacity:"))) {
                ptr += 25;
                /* FIX ? remain is float... */
                if (sscanf(ptr, "%d", &i->bat_active->remain) == 1) {
                    i->bat_active->percentage = ((i->bat_active->remain * 100)
                        / i->bat_active->currcap);
                } else {
                    /* fscanf failed */
                    i->bat_active->percentage = -1;
                }
            } else {
                /* when unconsistent data is met */
                i->bat_active->percentage = -1;
            }
        }
        free(file);
    } else {
        /* when the battery is not plugged in */
        i->bat_active->percentage = -1;
    }

    return retcode;
}

/*
 * Used with scandir to select "valid" device names. scandir would give the
 * full listing of the directory, including "." and "..", but we do not want
 * those two, so we need a custom dir_select function. (device names shouldn't
 * begin with a '.', so there shouldn't be any problem ).
 */
int dir_select (const struct dirent *f)
{
    return (f->d_name[0] == '.') ? 0 : 1;
}

/*
 * This function is called to parse a directory such as /proc/acpi/battery,
 * and to return both the number of files (*), and their names.
 *
 * If it cannot do it, 'list' is set to NULL, and it returns the error value
 * returned by scandir.
 * If scandir succeeds, the list of the file names is put into "list", and the
 * number of files (actually, those directories only contain other directories)
 * is returned.
 *
 * scandir and alphasort are not linux-specific.
 *
 * (*) Actually, these directories should only contain subdirectories, and no
 * regular file. But we do not check it for now.
 */
int parse_dir (char *dir, device_name **list)
{
    struct dirent **dirent_list;
    int i, n;

    n = scandir(dir, &dirent_list, dir_select, alphasort);

    if (n < 0) {
        fprintf(stderr, PACKAGE ": cannot parse directory %s : %s\n", dir,
                        strerror(errno));
        *list = NULL;
    } else {
        /* As we care only about the filenames, and not about all the other
         * informations contained in the dirent structure, we copy only what
         * we need into the namelist.
         */
        *list = (device_name *)malloc(sizeof(device_name)*n);

        if (*list == NULL) {
            fprintf(stderr, PACKAGE ": cannot allocate memory "
                            "for file list\n");
            perror("");
            exit(1);
        }

        for (i = 0; i < n; i++) {
            /*
             * d_name is a char[256] buffer, so is a device_name. So, using a
             * strncpy shouldn't be needed. Anyway, I prefer to do so, just in
             * case.
             */
            strncpy((*list)[i], dirent_list[i]->d_name,
                    sizeof(device_name) - 1);
            (*list)[i][sizeof(device_name) - 1] = '\0';
            free(dirent_list[i]);
        }
    }
    free(dirent_list);

    return n;
}

#endif  /* linux */

